(** Output signature for functor {!Bf.Make}. *)
module type BF = 
  sig
    type var = int
    type index
    and bf
	  
    (** maximal number of (unprimed) variables *) 
    val var_max : int

    (** reset variables *)
    val reset : unit -> unit	

    (** number of used variables *)
    val vars : unit -> int
	
    (** truth values *)
    val top      : bf
    val bottom   : bf
	
    (** operations on indices *)
    val var_index : var -> index
    val index_var : index -> var
    val prime_index : index -> index
    val unprime_index : index -> index

    (** variable constructors *)
    val new_var : unit -> var
    val mk_var : var -> bf
    val mk_primed_var : var -> bf
	
    (** boolean connectives *)
    val neg  : bf -> bf
    val conj : bf -> bf -> bf
    val disj : bf -> bf -> bf
    val impl : bf -> bf -> bf
    val equi : bf -> bf -> bf

    (** equality, ordering functions *)
    val eq       : bf -> bf -> bool
    val neq      : bf -> bf -> bool
    val le       : bf -> bf -> bool
    val ge       : bf -> bf -> bool
    (** val compare  : bf -> bf -> int *)
	
    (** relational product *)
    val rel_prod : bf -> bf -> bf	

    (** substitutions 
       val substitute_fn : (atom -> bf) -> bf -> bf
       val substitute : (atom * bf) list -> bf -> bf
     *)

    (** projection *)
    val exists : var list -> bf -> bf 
    val forall : var list -> bf -> bf 
	
    val foreach_cube : bf -> ((index -> int) -> unit) -> unit 

    val dom : bf -> var list
	
    val unprime_all : bf -> bf
    val prime_all : bf -> bf

    val dnf : bf -> index array list

    (* val size : bf -> int **)
    val to_string : (var -> string) -> bf -> string
    val print : out_channel -> (var -> string) -> bf -> unit
  end

module Bf : BF











