(** Interface to decision procedure/theorem prover *)

type dp_stats = {
    calls : int;
    cache_hits : int;
    total_time : float;
    max_time : float;
  }

module type DP = 
  sig
    type form
    val valid : Vcform.env -> Vcform.form -> bool
  end

module type DP_INTERFACE =
  sig
    (** check entailment - first argument is context *)
    val entails : Vcform.form -> Vcform.form -> Vcform.form -> bool
	
    (** check validity *)
    val valid : Vcform.form -> bool

    (** check satisfiability *)
    val satisfiable : Vcform.form -> bool
	
    (** report statistics *)
    val stats : unit -> dp_stats	
	
    (** reset statistics *)
    val reset_stats : unit -> unit

  end
     
module Make : functor (Dp : DP) -> DP_INTERFACE
    
module Dp : DP_INTERFACE
