{
 open Vcformparse
}
 let digitchar = ['0'-'9']
 let natlit = '-'? digitchar digitchar*
 let identchar = ['a'-'z''A'-'Z''_'''']
 let ident = identchar (identchar | digitchar)*
 let qualident = identchar (identchar | digitchar | '.')* identchar
 let xident = '\\' '<' identchar* '>'

  rule token = parse
  [' ' '\t''\n']     {token lexbuf}     (* skip blanks *)
| eof            {EOF}
| natlit as n    {NATLIT(int_of_string n)}
| "true"         {TRUE}
| "false"        {FALSE}
| "True"         {TRUE}
| "False"        {FALSE}
| "nullObj"      {NULL}
| "null"         {NULL}
| "{}"           {EMPTYSET}
| "\\<emptyset>" {EMPTYSET}
| "\\<forall>"    {FORALL}
| "ALL"          {FORALL}
| "\\<exists>"    {EXISTS}
| "EX"           {EXISTS}
| "lambda"       {LAMBDA}
| "%"            {LAMBDA}
| "Un"           {CUP}
| "\\<union>"     {CUP}
| "Int"          {CAP}
| "\\<inter>"     {CAP}
  (* constants *)
| "~="           {NEQ}
| "\\<noteq>"     {NEQ}
| '='            {EQ}
| '|'            {OR}
| "\\<or>"        {OR}
| '&'            {AND}
| "\\<and>"       {AND}
| '~'            {NOT}
| "\\<not>"       {NOT}
| "=>"           {IMPL}
| "-->"          {IMPL}
| "\\<longrightarrow>" {IMPL}
| "==>"          {METAIMPL}
| "<=>"          {IFF}
| '>'            {GT}
| '<'            {LT}
| ">="           {GTEQ}
| "<="           {LTEQ}
| "\\<le>"       {LTEQ}
| "[|"           {ALBRACKET}
| "|]"           {ARBRACKET}
| '['            {LBRACKET}
| ']'            {RBRACKET}
| '{'            {LBRACE}
| '}'            {RBRACE}
| ','            {COMMA}
| ".."           {DEREF}
| '.'            {DOT}
| '('            {LPAREN}
| ')'            {RPAREN}
| '+'            {PLUS}
| '-'            {MINUS}
| '*'            {TIMES}
| "div"          {DIVIDE}
| "::"           {COLONCOLON}
| ':'            {COLON}
| "\\<in>"       {COLON}
| "~:"           {NCOLON}
| "\\<notin>"    {NCOLON}
| '@'            {APPLY}
| ';'            {SEMICOLON}
| "arrayRead"    {ARRAYREAD}
| "arrayWrite"   {ARRAYWRITE}
| "newArray"     {NEWARRAY}
| "arraySize"    {ARRAYSIZE}
| "fieldWrite"   {FIELDWRITE}
| "fieldRead"    {FIELDREAD}
| "bool"         {BOOL}
| "int"          {INT}
| "ref"          {REF}
| "array"        {ARRAY}
| "set"          {SET}
| "list"         {LIST}
| "void"         {VOID}
| "universe"     {UNIVERSE}
| "->"           {ARROW}
| ident as s     {IDENT s}
| qualident as s {QIDENT s}
| xident as s    {failwith ("unknown x-symbol token '" ^ s ^ "'")}
