/* note: canonical version in tests/httpd */

scope HDLLScope {
    modules HDLLIter; 
    exports HDLLIter;
    invariant HDLLIter.Iter in HDLLIter.Content;
}

spec module HDLLIter {
  format Node;

  specvar init : bool;
  specvar Content : Node set;
  specvar Iter : Node set;

  default I = init;

  proc init()
    suspends I
    requires not init
    modifies init
    ensures card(Content)=0 & init';

  proc clear() requires init & card(Content) >= 1 & (Iter in Content)
    modifies Content, Iter
    ensures card(Content') = 0 & card(Iter') = 0;

  proc add(n : Node)
    requires card(n)=1 & not (n in Content)
    modifies Content
    ensures (Content' = Content + n);

  proc remove(n : Node)
    requires card(n)=1 & (n in Content)
    modifies Content, Iter
    ensures (Content' = Content - n) &
            (Iter' = Iter - n);

  proc removeFirst() returns n:Node
    requires card(Content)>=1
    modifies Content, Iter
    ensures card(n')=1 & (Content' = Content - n') & (n' in Content) &
            (Iter' = Iter - n');

  proc isEmpty() returns rv:bool
    ensures rv' <=> (card(Content) = 0 & card(Iter) = 0);

  proc openIter()
    requires card(Iter) = 0
    modifies Iter
    ensures (Iter' = Content);

  proc nextIter() returns n : Node
    requires card(Iter)>=1
    modifies Iter
    ensures card(n')=1 & (n' in Iter) & 
            (Iter' = Iter - n');

  proc isLastIter() returns e:bool 
    ensures not e' <=> (card(Iter') >= 1);

  proc closeIter()
    modifies Iter
    ensures card(Iter') = 0;
}
