spec module Compiler
{
    // EXPR_STMT   = 0x001
    // ASSIGN_EXPR = 0x100
    // ADD_EXPR    = 0x101
    // LOCAL_VAR   = 0x102
    // INT_CONST   = 0x200

    format Node;

    specvar Transformed, Untransformed : Node set;
    // also use TransformedList.InList, UntransformedList.InList

    predicate @repinv () = 
                 (Transformed = TransformedList.InList) &
                 (Untransformed = UntransformedList.InList) &
                 disjoint (UntransformedList.InList, TransformedList.InList);

    proc markTransformed(n : Node)
        requires (n in UntransformedList.InList) & card(n)=1 &
                 @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList, TransformedList
        ensures (UntransformedList.InList' = UntransformedList.InList - n) &
                (TransformedList.InList' = TransformedList.InList + n) &
                @repinv' ();

    // foldConstants guarantees that:
    // 1) all typestate flags 'transformed' changed from 1 to 2
              // currently have partial result: n changes 1 to 2
    // 2) all objects are always in at most one list and at most one tree,
              // done, using disjointness, except for reachability
    // 3) new tree&list contain nodes from the old list plus some new nodes.

    proc foldConstants (n : Node) returns r : Node
        requires (n in Untransformed) & card(n)=1 &
                 @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        ensures @repinv' ();

    proc newExprStmt (e : Node) returns r : Node
        requires @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList
        ensures (Untransformed' = Untransformed + r') &
                @repinv' ();

    proc newAssignExpr (l : Node; rhs : Node) returns r : Node
        requires @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList
        ensures (Untransformed' = Untransformed + r') &
                @repinv' ();

    proc newAddExpr (l : Node; rhs : Node) returns r : Node
        requires @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList
        ensures @repinv' ();

    proc newLocalVarExpr (v : string) returns r : Node
        requires @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList
        ensures (Untransformed' = Untransformed + r') &
                @repinv' ();

    proc newIntConst (i : int) returns r : Node
        requires @repinv ()
        modifies UntransformedList.InList, TransformedList.InList,
                 Transformed, Untransformed
        calls UntransformedList
        ensures (Untransformed' = Untransformed + r') &
                @repinv' ();
}
