spec module Domain
{
    /* --- measurements --- */
    format Measurement;

    format TempM;
    specvar TM_Failed : TempM set; // 255
    specvar TM_TooHot : TempM set;
    specvar TM_TooCold : TempM set;
    specvar TM_JustRight : TempM set;
    proc TM_Failed (m:Measurement) returns rv:bool
        ensures rv' <=> m in TM_Failed;
    proc TM_TooHot (m:Measurement) returns rv:bool
        ensures rv' <=> m in TM_TooHot;
    proc TM_TooCold (m:Measurement) returns rv:bool
        ensures rv' <=> m in TM_TooCold;
    proc TM_JustRight (m:Measurement) returns rv:bool
        ensures rv' <=> m in TM_JustRight;

    /* --- state variables --- */

    format TempSV;
    specvar Temp_Unknown : TempSV set;
    specvar Temp_TooHot : TempSV set;
    specvar Temp_TooCold : TempSV set;
    specvar Temp_JustRight : TempSV set;
    proc Temp_setUnknown(s:State) 
	ensures s in Temp_Unknown' & 
	        not (s in (Temp_TooHot' + Temp_TooCold' + Temp_JustRight'));
    proc Temp_setTooHot(s:State) 
	ensures s in Temp_TooHot' & 
	        not (s in (Temp_Unknown' + Temp_TooCold' + Temp_JustRight'));
    proc Temp_setTooCold(s:State) 
	ensures s in Temp_TooCold' & 
	        not (s in (Temp_TooHot' + Temp_Unknown' + Temp_JustRight'));
    proc Temp_setJustRight(s:State) 
	ensures s in Temp_JustRight' & 
	        not (s in (Temp_TooHot' + Temp_TooCold' + Temp_Unknown'));

    format SensorHealthSV;
    specvar SH_Unknown : SensorHealthSV set; // never reached...
    specvar SH_Healthy : SensorHealthSV set;
    specvar SH_Failed : SensorHealthSV set;
    proc SH_isHealthy(s:State) returns rv:bool ensures rv' <=> s in SH_Healthy;
    proc SH_setFailed(s:State) 
	modifies SH_Unknown, SH_Healthy, SH_Failed 
	ensures s in SH_Failed' & 
		not (s in (SH_Unknown' + SH_Healthy'));
    proc SH_setHealthy(s:State) 
	modifies SH_Unknown, SH_Healthy, SH_Failed 
	ensures s in SH_Healthy' & 
		not (s in (SH_Unknown' + SH_Failed'));

    format HeaterSwitchSV;
    specvar HS_Unknown : HeaterSwitchSV set;
    specvar HS_Healthy : HeaterSwitchSV set;
    specvar HS_Stuck : HeaterSwitchSV set;
    specvar HS_Open : HeaterSwitchSV set;
    specvar HS_Closed : HeaterSwitchSV set;

    format State;
    proc setController1Command(s:State, c:SwitchControllerCommand) 
        ensures true;
    proc setController2Command(s:State, c:SwitchControllerCommand) 
        ensures true;

    /* --- commands --- */

    format SwitchControllerCmd;
    specvar SC_Unknown : SwitchControllerCmd set;
    specvar SC_Open : SwitchControllerCmd set;
    specvar SC_Closed : SwitchControllerCmd set;

    proc SC_setUnknown (c:SwitchControllerCmd)
	modifies SC_Unknown, SC_Open, SC_Closed 
	ensures s in SC_Unknown' & 
		not (s in (SC_Open' + SC_Closed'));
    proc SC_setOpen (c:SwitchControllerCmd)
	modifies SC_Unknown, SC_Open, SC_Closed 
	ensures s in SC_Open' & 
		not (s in (SC_Unknown' + SC_Closed'));
    proc SC_setClosed (c:SwitchControllerCmd)
	modifies SC_Unknown, SC_Open, SC_Closed 
	ensures s in SC_Closed' & 
		not (s in (SC_Open' + SC_Unknown'));
}
