spec module RemainingEstimators = List with Est <- Entry;

spec module Estimators
{
    format Est;

    /* we can only run an instance of an estimator once currently. */

    proc init() ensures true;
    proc nextCycle() ensures true;
    proc estimatorsUnconsumed() returns rv : bool
        ensures (not rv') <=> card(RemainingEstimators.Content) = 0;
}

spec module TempEstimator
{
    format Est; format State; format Measurement;
    specvar init : bool;
    specvar THIS : Est set;
    default T = card(THIS') = 1;

    proc init() suspends T modifies init, THIS ensures init' & card(THIS') = 1;
    proc nextCycle() 
        requires init
        modifies RemainingEstimators.Content 
        ensures THIS in RemainingEstimators.Content';

    proc estimate(s:State; m:Measurement) 
        modifies RemainingEstimators.Content 
        ensures (RemainingEstimators.Content' = RemainingEstimators.Content - THIS);
}

spec module SensorHealthEstimator
{
    format Est; format State; format Measurement;
    specvar init : bool;
    specvar THIS : Est set;
    proc init() modifies init, THIS ensures init' & card(THIS') = 1;
    proc nextCycle() 
        requires init
        modifies RemainingEstimators.Content 
        ensures THIS in RemainingEstimators.Content';

    proc estimate(s:State; m:Measurement) 
        modifies RemainingEstimators.Content 
        ensures card(THIS') = 1 & (RemainingEstimators.Content' = RemainingEstimators.Content - THIS);
}

spec module SwitchEstimator 
{
    format Est;
    specvar init : bool;
    specvar THIS : Est set;
    proc init() modifies init, THIS ensures init' & card(THIS') = 1;
    proc nextCycle() 
        requires init
        modifies RemainingEstimators.Content 
        ensures THIS in RemainingEstimators.Content';

    proc estimate(s:State) 
        requires not (TempEstimator.THIS in RemainingEstimators.Content) &
                 not (SensorHealthEstimator.THIS in RemainingEstimators.Content)
        modifies RemainingEstimators.Content 
        ensures RemainingEstimators.Content' = RemainingEstimators.Content - THIS;
}
