/* main control loop */

/* spec module EventQueue = List with Event <- Entry; */

spec module Simulator
{
    format State;
    format Measurement;
    format SwitchControllerCommand;

    default sNotNull(f, s) = f & card(s) = 1;

    specvar INITIAL:State set;
    specvar INPUT:State set;
    specvar ESTIMATE:State set; 
    specvar OUTPUT:State set;
    specvar EXECUTE:State set;
    specvar DONE:State set;

    proc simulateOnce() 
        requires card(INITIAL) = 0 & card(INPUT) = 0 &
                 card(ESTIMATE) = 0 & card(OUTPUT) = 0 &
		 card(EXECUTE) = 0 & card(DONE) = 0
        modifies INITIAL, INPUT, ESTIMATE, OUTPUT, 
		 EXECUTE, DONE, RemainingEstimators.Content,
		 Domain.SC_Unknown, Domain.SC_Open, Domain.SC_Closed
        ensures card(DONE') = 1;

    proc procureInput(s:State) returns m:Measurement
        requires (INPUT = s) & card(ESTIMATE)=0
        modifies INPUT, ESTIMATE
	calls SensorHWA
        ensures card(INPUT')=0 & (ESTIMATE'=s) & card(m') = 1;

    proc acceptInput(s:State; m:Measurement) 
        requires (ESTIMATE = s) & (card(m) = 1) & (card(OUTPUT) = 0)
        modifies ESTIMATE, OUTPUT, RemainingEstimators.Content
	calls Estimators, TempEstimator, SensorHealthEstimator, SwitchEstimator, RemainingEstimators
        ensures (OUTPUT' = s) & card(ESTIMATE') = 0 &
                (card(RemainingEstimators.Content') = 0);

    proc produceOutput(s:State)
        requires (OUTPUT = s) & card(EXECUTE) = 0 &
	         (card(RemainingEstimators.Content) = 0);
        modifies OUTPUT, EXECUTE, Domain.SC_Unknown, Domain.SC_Open, Domain.SC_Closed
	calls SwitchController
        ensures (EXECUTE' = s) & card(OUTPUT') = 0;

    proc executeCommand(s:State)
        requires (EXECUTE = s) & card(DONE) = 0
        modifies EXECUTE, DONE
        calls SwitchHWA
        ensures (DONE' = s) & card(EXECUTE') = 0;
}
