spec module CacheSet = HDLLIter with Node <- Entry;
spec module CacheBlacklist = HDLLIter with Node <- Entry;

scope CacheScope {
    modules Cache, CacheSet, CacheBlacklist;
    exports Cache;
    invariant Cache.init => CacheSet.init & CacheBlacklist.init & card(CacheSet.Iter) = 0 & card(CacheBlacklist.Iter) = 0;
}

spec module Cache {
    format Entry;
    default I = Cache.init;
    specvar init:bool;

    proc init() suspends I 
        requires not CacheSet.init & not CacheBlacklist.init
        modifies init, CacheSet.init, CacheBlacklist.init, CacheSet.Iter, CacheBlacklist.Iter
        ensures init';

    proc setEntryName (n:Entry; fn:string) ensures true;
    proc setEntryModTime (n:Entry; modTime:float) ensures true;
    proc setEntryFlags (n:Entry; f:string) ensures true;
    proc setEntryContent (n:Entry; s:byte[]) ensures true;

    proc getEntryName (n:Entry) returns fn:string ensures true;
    proc getEntryModTime (n:Entry) returns modTime:float ensures true;
    proc getEntryFlags (n:Entry) returns f:string ensures true;
    proc getEntryContent (n:Entry) returns s:byte[] ensures true;

    proc addBlacklistedFile (n:Entry) 
        requires card(n) = 1
        modifies CacheBlacklist.Content
        ensures CacheBlacklist.Content' = CacheBlacklist.Content + n;

    proc removeBlacklistedFile (n:Entry) 
        requires card(n) = 1
        modifies CacheBlacklist.Content
        ensures not (n in CacheBlacklist.Content');

    proc hasEntry (n:Entry) returns rv:bool
        ensures rv' <=> (n in CacheSet.Content); // Object identity

    proc removeEntry (n:Entry)
        requires card(n) = 1
        ensures not (n in CacheSet.Content');

    proc addEntry (n:Entry) 
        requires card(n) = 1 & not (n in CacheSet.Content)
        modifies CacheSet.Content
        ensures n in CacheSet.Content';

    proc loadEntryContent (n:Entry)
	requires card(n) = 1 /*& (n in CacheSet.Content)*/
        ensures true;

    proc getLength (n:Entry) returns c:int
	requires card(n) = 1
	ensures true;
}
