spec module Sendfile {
    default I = HTTPRequest.init & Mimetypes.init & Cache.init;

    private proc badFileName(s:string) returns rv:bool ensures true;
    private proc checkNotModified (tmMod:float; modTime:string) returns rv:bool ensures true;
    private proc ce (n:string; f:string; lastmod:float; content:byte[]) returns h:Entry ensures card(h') = 1;
    private proc he (k:string; v:string) returns h:Entry ensures card(h') = 1;
    private proc checkCanZip () returns rv:bool requires card(StringTokenizer.S) = 0 modifies StringTokenizer.S ensures true;
    private proc zipEntry (c:Entry; cz:Entry) 
      requires card(cz) = 1 & card(c) = 1
      modifies HTTPResponse.C, CacheSet.Content
      ensures cz' in CacheSet.Content';
    private proc prepHead (oc:out_channel; name:string; tmMod:float) 
      modifies HTTPResponse.C
      ensures card(HTTPResponse.C') >= 1;
    private proc putLeShort (targ:byte[]; off:int; n:int) ensures true;
    private proc putLeInt (targ:byte[]; off:int; n:int) ensures true;
    private proc sendEntry (oc:out_channel; n:Entry) returns c:int
      requires (n in CacheSet.Content) | (n in CacheBlacklist.Content) 
      ensures true;

    format retval;
    proc getStatus(rv:retval) returns s:int ensures true;
    proc getCount(rv:retval) returns c:int ensures true;

    proc OK() returns r:int ensures true;
    proc REDIRECTION() returns r:int ensures true;
    proc NOT_MODIFIED() returns r:int ensures true;
    proc BAD_REQUEST() returns r:int ensures true;
    proc FORBIDDEN() returns r:int ensures true;
    proc NOT_FOUND() returns r:int ensures true;
    proc r_OK(count:int) returns rv:retval ensures true;
    proc r_NOT_MODIFIED() returns rv:retval ensures true;
    proc r_FORBIDDEN() returns rv:retval ensures true;
    proc r_NOT_FOUND() returns rv:retval ensures true;

    // returns neg numbers for errors, positive numbers = bytes sent
    proc sendFile (out:file_descr; name:string; modTime:string;
                   sendHead:bool; sendBody:bool) returns rv:retval
      requires Mimetypes.init & card(StringTokenizer.S) = 0 & 
               card(HTTPResponse.C) = 0
      modifies StringTokenizer.S, HTTPResponse.C, CacheBlacklist.Content, CacheSet.Content
      calls Util
      ensures card(HTTPResponse.C') = 0;
}
