spec module Water
{
    format MoleculeData;
    format ForceData;
    format Energy;

    specvar Corrected, Predicted : MoleculeData set;
    specvar Computed, Consumed : Energy set;
    specvar InterfUncomputed, InterfComputed : ForceData set;
    specvar IntrafUncomputed, IntrafComputed : ForceData set;

    proc INITIAM() returns c : MoleculeData
        requires disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Corrected
        ensures (c' in Corrected') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc INITIAF() returns c : ForceData
        requires disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies InterfUncomputed, IntrafUncomputed
        ensures (c' in InterfUncomputed') & 
                (c' in IntrafUncomputed') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc PREDIC(c:MoleculeData) returns p : MoleculeData
        requires (c in Corrected) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Corrected, Predicted
        ensures (p' = c) & (Predicted' = Predicted + p') &
                (Corrected' = Corrected - p') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc INTRAF(m:MoleculeData; f:ForceData) returns cf : ForceData
        requires (m in Predicted) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies IntrafUncomputed, IntrafComputed
        ensures (cf' = f) & 
                (IntrafUncomputed' = IntrafUncomputed - cf') &
                (IntrafComputed' = IntrafComputed + cf') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc INTERF(m:MoleculeData; f:ForceData) returns cf : ForceData
        requires (m in Predicted) & (f in InterfUncomputed) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies InterfUncomputed, InterfComputed
        ensures (cf' = f) & 
                (InterfUncomputed' = InterfUncomputed - cf') &
                (InterfComputed' = InterfComputed + cf') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc CORREC(m:MoleculeData; cf:ForceData) returns c : MoleculeData
        requires (m in Predicted) & (cf in InterfComputed) & 
                 (cf in IntrafComputed) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Predicted, Corrected, InterfComputed, InterfUncomputed,
                 IntrafComputed, IntrafUncomputed
        ensures (c' = m) &
                (Predicted' = Predicted - c') &
                (Corrected' = Corrected + c') &
                (InterfComputed' = InterfComputed - cf) &
                (InterfUncomputed' = InterfUncomputed + cf) &
                (IntrafComputed' = IntrafComputed - cf) &
                (IntrafUncomputed' = IntrafUncomputed + cf) &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc KINETI(m:MoleculeData) returns k:Energy
        requires (m in Corrected) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Computed
        ensures (Computed' = Computed + k') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc POTENG(m:MoleculeData) returns p:Energy
        requires (m in Corrected) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Computed
        ensures (Computed' = Computed + p') &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc OUTPUT(k:Energy; p:Energy)
        requires (k in Computed) & (p in Computed) &
                 disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Computed, Consumed
        ensures (Computed' = Computed - (k + p)) &
                (Consumed' = Consumed + (k + p)) &
                disjoint(Corrected',Predicted') & 
                disjoint(Computed',Consumed') &
                disjoint(InterfUncomputed',InterfComputed') &
                disjoint(IntrafUncomputed',IntrafComputed');

    proc mainLoop() 
        requires disjoint(Corrected,Predicted) & disjoint(Computed,Consumed) &
                 disjoint(InterfUncomputed,InterfComputed) &
                 disjoint(IntrafUncomputed,IntrafComputed)
        modifies Computed, Consumed, Corrected, Predicted, InterfUncomputed,
                 IntrafUncomputed, IntrafComputed, InterfComputed
        ensures true;
}
