spec module SuspendedQueue = PriorityQueue with PQEntry <- Process;
spec module RunningList = DoublyLinkedList with Node <- Process;

spec module Scheduler 
{
    format Process;
    specvar Running, Suspended : Process set;

    predicate @inv() = (Running = RunningList.InList) &
                       (Suspended = SuspendedQueue.PQ) &
                       disjoint (Running, Suspended);

    proc suspend(p : Process; priority : int)
        requires (p in Running) & card(p)=1 & @inv()
        modifies Running, Suspended, RunningList.InList, SuspendedQueue.PQ
        calls RunningList, SuspendedQueue
        ensures (Suspended' = Suspended + p) &
                (Running' = Running - p) &
                @inv'();

    proc hasSuspended() returns b : bool
        requires @inv()
        calls SuspendedQueue
        ensures b' <=> (Suspended != {});

    proc wakeUpFirst()
         requires (Suspended != {}) & @inv()
         modifies Running, Suspended, RunningList.InList, SuspendedQueue.PQ
         calls SuspendedQueue, RunningList 
         ensures exists r : Process.
                 ((r in Suspended) &
                 (Suspended' = Suspended - r)  &
                 (Running' = Running + r)) &
                 @inv'();
}
